/*\
* lenactions
* A simple image editing library
* Copyrigth 2013 - Hadrien Croubois
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
\*/

#include <lenactions/utils.hh>

template<typename T>
lenactions::image<T>::image()
	: m_rows(0)
	, m_cols(0)
{
}

template<typename T>
lenactions::image<T>::image(std::istream& in)
{
	this->from_stream(in);
}

template<typename T>
lenactions::pixel<T>&
lenactions::image<T>::operator[] (int idx)
{
	return m_bitmap[idx];
}

template<typename T>
const lenactions::pixel<T>&
lenactions::image<T>::operator[] (int idx) const
{
	return m_bitmap[idx];
}

template<typename T>
std::pair<size_t, size_t>
lenactions::image<T>::dimensions() const
{
	return std::make_pair(m_cols, m_rows);
}

template<typename T>
void
lenactions::image<T>::from_stream(std::istream& in)
{
	switch (lenactions::tools::getc(in))
	{
		case 'P': break;
		default:  throw;
	}

	format in_format = (format) lenactions::tools::getint(in);
	this->m_cols = lenactions::tools::getint(in);
	this->m_rows = lenactions::tools::getint(in);
	this->m_bitmap.resize(this->m_cols*this->m_rows);
	int scale = (in_format == P1  || in_format == P4)?1:lenactions::tools::getint(in);

	for (lenactions::pixel<T>& p : this->m_bitmap)
	{
		switch (in_format)
		{
			case 1:
			{
				p = pixel<T>((T) !lenactions::tools::getint(in));
				break;
			}
			case 2:
			{
				p = pixel<T>((T) lenactions::tools::getint(in)/scale);
				break;
			}
			case 3:
			{
				T r = (T) lenactions::tools::getint(in)/scale;
				T g = (T) lenactions::tools::getint(in)/scale;
				T b = (T) lenactions::tools::getint(in)/scale;
				p = lenactions::pixel<T>({r, g, b}, RGB);
				break;
			}
			case 4:
			{
				p = pixel<T>((T) !lenactions::tools::getrawbyte(in));
				break;
			}
			case 5:
			{
				p = pixel<T>((T) lenactions::tools::getrawbyte(in)/scale);
				break;
			}
			case 6:
			{
				T r = (T) lenactions::tools::getrawbyte(in)/scale;
				T g = (T) lenactions::tools::getrawbyte(in)/scale;
				T b = (T) lenactions::tools::getrawbyte(in)/scale;
				p = lenactions::pixel<T>({r, g, b}, RGB);
				break;
			}
		}
	}
}

template<typename T>
void
lenactions::image<T>::to_stream(std::ostream& out, format out_format)
{
	out << "P" << out_format << std::endl;
	out << this->m_cols << " " << this->m_rows << std::endl;
	switch (out_format)
	{
		case P1:
		case P4:
			break;
		case P2:
		case P3:
		case P5:
		case P6:
			out << 255 << std::endl;
	}
	for (lenactions::pixel<T>& p : this->m_bitmap)
		p.to_stream(out, out_format);
}

// ┌──────────────────────────────────────────────────────────────────────────┐
// │                                  TP 02                                   │
// └──────────────────────────────────────────────────────────────────────────┘
template <typename T>
lenactions::image<T>&
lenactions::image<T>::histogram(std::ostream& out)
{
	std::array<std::array<size_t, 256>, 4> cdf = { 0 };

	for (pixel<T>& p : this->m_bitmap)
	{
		++cdf[0][(size_t) (255 * p.r())];
		++cdf[1][(size_t) (255 * p.g())];
		++cdf[2][(size_t) (255 * p.b())];
		++cdf[3][(size_t) (255 * p.v())];
	}
	for (size_t i=0; i<256; ++i)
	{
		out << i << " "
				<< (double) cdf[0][i] / (this->m_rows*this->m_cols) << " "
				<< (double) cdf[1][i] / (this->m_rows*this->m_cols) << " "
				<< (double) cdf[2][i] / (this->m_rows*this->m_cols) << " "
				<< (double) cdf[3][i] / (this->m_rows*this->m_cols) << std::endl;
	}
	return *this;
}

template <typename T>
lenactions::image<T>&
lenactions::image<T>::gamma(T y)
{
	for (pixel<T>& p : this->m_bitmap) p.v() = pow(p.v(), 1/y);
	return *this;
}

template <typename T>
lenactions::image<T>&
lenactions::image<T>::rotate(T y)
{
	for (pixel<T>& p : this->m_bitmap) p.h() += y;
	return *this;
}

template <typename T>
lenactions::image<T>&
lenactions::image<T>::normalize()
{
	std::array<T, 256> cdf = { T() };

	for (pixel<T>& p : this->m_bitmap)
		++cdf[(size_t) (255 * p.v())];
	for (size_t i=1; i<256; ++i)
		cdf[i] += cdf[i-1];

	T min = cdf[0], max = cdf[255];

	cdf[0] = 0;
	for (size_t i=1; i<256; ++i)
		cdf[i] = (cdf[i] - min) / (max - min);

	for (pixel<T>& p : this->m_bitmap)
		p.v() = cdf[(size_t) (255 * p.v())];

	return *this;
}

// ┌──────────────────────────────────────────────────────────────────────────┐
// │                                  TP 03                                   │
// └──────────────────────────────────────────────────────────────────────────┘

template <typename T>
template <size_t H, size_t V>
lenactions::image<T>&
lenactions::image<T>::compose(const lenactions::Kernel<T,H,V>& kernel, const T& norm)
{
	std::vector<pixel<T>> new_bitmap(this->m_bitmap.size(), T(0));

	for (size_t i=0; i<this->m_rows; ++i)
		for (size_t j=0; j<this->m_cols; ++j)
		{
			pixel<T>& p = new_bitmap[i*this->m_cols+j];
			for (uint64_t k=0; k<H; ++k)
				for (uint64_t l=0; l<V; ++l)
				{
					int64_t kk = (int64_t) k - H / 2;
					int64_t ll = (int64_t) l - H / 2;
					p.r() += kernel[k][l] * this->m_bitmap[o_rows(i,kk) * this->m_cols + o_cols(j,ll)].r();
					p.g() += kernel[k][l] * this->m_bitmap[o_rows(i,kk) * this->m_cols + o_cols(j,ll)].g();
					p.b() += kernel[k][l] * this->m_bitmap[o_rows(i,kk) * this->m_cols + o_cols(j,ll)].b();
				}
			p.r() /= norm;
			p.g() /= norm;
			p.b() /= norm;
		}
	this->m_bitmap = std::move(new_bitmap);
	return *this;
}

template <typename T>
lenactions::image<T>
lenactions::image<T>::assemblage(image<T>& a, image<T>& b, const pixelOperator<T>& op)
{
	if (a.dimensions() != b.dimensions()) throw;

	lenactions::image<T> assembled;
	assembled.m_rows   = a.m_rows;
	assembled.m_cols   = a.m_cols;
	assembled.m_bitmap.resize(a.m_rows*a.m_cols);

	for (size_t i=0; i<a.m_rows*a.m_cols; ++i)
		assembled.m_bitmap[i] = op(a.m_bitmap[i], b.m_bitmap[i]);

	return assembled;
}

// ────────────────────────────────────────────────────────────────────────────

template <typename T>
lenactions::image<T>&
lenactions::image<T>::smooth()
{
	this->compose<1,3>({{T(1), T(2), T(1)}}, 4.0).compose<3,1>({{T(1), T(2), T(1)}}, 4.0);
	return *this;
}

template <typename T>
lenactions::image<T>&
lenactions::image<T>::contour_prewitt(const pixelOperator<T>& op)
{
	lenactions::image<T> cpy_h = lenactions::image<T>(*this).compose<1,3>({{T(-1), T(+0), T(+1)}}, T(1)).compose<3,1>({{T(1), T(1), T(1)}}, T(3));
	lenactions::image<T> cpy_v = lenactions::image<T>(*this).compose<3,1>({{T(-1), T(+0), T(+1)}}, T(1)).compose<1,3>({{T(1), T(1), T(1)}}, T(3));
	for (size_t i=0; i<m_rows*m_cols; ++i) this->m_bitmap[i] = op(cpy_h.m_bitmap[i], cpy_v.m_bitmap[i]);
	// *this = lenactions::image<T>::assemblage(cpy_h, cpy_v, op);
	return *this;
}

template <typename T>
lenactions::image<T>&
lenactions::image<T>::contour_sobel(const pixelOperator<T>& op)
{
	lenactions::image<T> cpy_h = lenactions::image<T>(*this).compose<1,3>({{T(-1), T(+0), T(+1)}}, T(1)).compose<3,1>({{T(1), T(2), T(1)}}, T(3));
	lenactions::image<T> cpy_v = lenactions::image<T>(*this).compose<3,1>({{T(-1), T(+0), T(+1)}}, T(1)).compose<1,3>({{T(1), T(2), T(1)}}, T(3));
	for (size_t i=0; i<m_rows*m_cols; ++i) this->m_bitmap[i] = op(cpy_h.m_bitmap[i], cpy_v.m_bitmap[i]);
	// *this = lenactions::image<T>::assemblage(cpy_h, cpy_v, op	);
	return *this;
}

template <typename T>
lenactions::image<T>&
lenactions::image<T>::contour_kirsch()
{
	lenactions::image<T> cpy_NN = lenactions::image<T>(*this).compose<3,3>(std::array<std::array<T, 3>,3>({{ {+5, +5, +5}, {-3, +0, -3}, {-3, -3, -3} }}));
	lenactions::image<T> cpy_NE = lenactions::image<T>(*this).compose<3,3>(std::array<std::array<T, 3>,3>({{ {-3, +5, +5}, {-3, +0, +5}, {-3, -3, -3} }}));
	lenactions::image<T> cpy_EE = lenactions::image<T>(*this).compose<3,3>(std::array<std::array<T, 3>,3>({{ {-3, -3, +5}, {-3, +0, +5}, {-3, -3, +5} }}));
	lenactions::image<T> cpy_SE = lenactions::image<T>(*this).compose<3,3>(std::array<std::array<T, 3>,3>({{ {-3, -3, -3}, {-3, +0, +5}, {-3, +5, +5} }}));
	lenactions::image<T> cpy_SS = lenactions::image<T>(*this).compose<3,3>(std::array<std::array<T, 3>,3>({{ {-3, -3, -3}, {-3, +0, -3}, {+5, +5, +5} }}));
	lenactions::image<T> cpy_SW = lenactions::image<T>(*this).compose<3,3>(std::array<std::array<T, 3>,3>({{ {-3, -3, -3}, {+5, +0, -3}, {+5, +5, -3} }}));
	lenactions::image<T> cpy_WW = lenactions::image<T>(*this).compose<3,3>(std::array<std::array<T, 3>,3>({{ {+5, -3, -3}, {+5, +0, -3}, {+5, -3, -3} }}));
	lenactions::image<T> cpy_NW = lenactions::image<T>(*this).compose<3,3>(std::array<std::array<T, 3>,3>({{ {+5, +5, -3}, {+5, +0, -3}, {-3, -3, -3} }}));
	for (size_t i=0; i<m_rows*m_cols; ++i)
	{
		this->m_bitmap[i].r() = std::max({ cpy_NN[i].r(), cpy_NE[i].r(), cpy_EE[i].r(), cpy_SE[i].r(), cpy_SS[i].r(), cpy_SW[i].r(), cpy_WW[i].r(), cpy_NW[i].r() });
		this->m_bitmap[i].g() = std::max({ cpy_NN[i].g(), cpy_NE[i].g(), cpy_EE[i].g(), cpy_SE[i].g(), cpy_SS[i].g(), cpy_SW[i].g(), cpy_WW[i].g(), cpy_NW[i].g() });
		this->m_bitmap[i].b() = std::max({ cpy_NN[i].b(), cpy_NE[i].b(), cpy_EE[i].b(), cpy_SE[i].b(), cpy_SS[i].b(), cpy_SW[i].b(), cpy_WW[i].b(), cpy_NW[i].b() });
	//this->m_bitmap[i].h() = 0;
	//this->m_bitmap[i].s() = 0;
	//this->m_bitmap[i].v() = std::max({ cpy_NN[i].v(), cpy_NE[i].v(), cpy_EE[i].v(), cpy_SE[i].v(), cpy_SS[i].v(), cpy_SW[i].v(), cpy_WW[i].v(), cpy_NW[i].v() });
	}
	return *this;
}

// ┌──────────────────────────────────────────────────────────────────────────┐
// │                                 PRIVATE                                  │
// └──────────────────────────────────────────────────────────────────────────┘

template <typename T>
size_t
lenactions::image<T>::o_cols(int64_t j, int64_t offset) const
{
	int64_t r = j+offset;
	while (r < 0 || r >= (int64_t) this->m_cols)
		r = (r < 0)?(-r):(2*(this->m_cols-1)-r);
	return r;
}

template <typename T>
size_t
lenactions::image<T>::o_rows(int64_t i, int64_t offset) const
{
	int64_t r = i+offset;
	while (r < 0 || r >= (int64_t) this->m_rows)
		r = (r < 0)?(-r):(2*(this->m_rows-1)-r);
	return r;
}
