/*\
* lenactions
* A simple image editing library
* Copyrigth 2013 - Hadrien Croubois
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
\*/

template<typename T>
lenactions::pixel<T>::pixel()
	: m_space(RGB)
	, m_data({T(), T(), T()})
{
}

template<typename T>
lenactions::pixel<T>::pixel(const T& g)
	: m_space(RGB)
	, m_data({g, g, g})
{
}

template<typename T>
lenactions::pixel<T>::pixel(std::array<T,3> colors, colorSpace space)
  : m_space(space)
  , m_data(colors)
{
}

template<typename T>
T&
lenactions::pixel<T>::get_canal(canal c)
{
	switch (c)
	{
		case R:
		case G:
		case B:
			this->set_space(RGB);
			return this->m_data[c];
		case H:
		case S:
		case V:
			this->set_space(HSV);
			return this->m_data[c-3];
		case GREY:
			this->set_space(HSV);
			return this->m_data[2]; // V
	}
	throw;
}

template<typename T>
void
lenactions::pixel<T>::set_canal(canal c, const T& v)
{
	switch (c)
	{
		case R:
		case G:
		case B:
			this->set_space(RGB);
			this->m_data[c] = v;
			break;
		case H:
		case S:
		case V:
			this->set_space(HSV);
			this->m_data[c-3] = v;
			break;
		case GREY:
			if (this->m_space == HSV)
			{
				this->m_data[1] = 0;
				this->m_data[2] = v;
			}
			else
			{
				this->m_data[0] = v;
				this->m_data[1] = v;
				this->m_data[2] = v;
			}
	}
}

template<typename T>
void
lenactions::pixel<T>::set_space(colorSpace s)
{
	if (this->m_space == s) return;
	switch (s)
	{
		case RGB:
		{
			T r;
			T g;
			T b;
			T h = this->h() / 60;
			T s = this->s();
			T v = this->v();
			T c = s * v;
			T x = c * (1.0 - std::abs(fmod(h, 2.0) - 1.0));

			switch ((int) h % 6)
			{
				case 0:
					r = c;
					g = x;
					b = 0;
					break;
				case 1:
					r = x;
					g = c;
					b = 0;
					break;
				case 2:
					r = 0;
					g = c;
					b = x;
					break;
				case 3:
					r = 0;
					g = x;
					b = c;
					break;
				case 4:
					r = x;
					g = 0;
					b = c;
					break;
				case 5:
					r = c;
					g = 0;
					b = x;
					break;
				default:
					throw;
			}
			this->m_space   = RGB;
			this->m_data[0] = r+v-c;
			this->m_data[1] = g+v-c;
			this->m_data[2] = b+v-c;
			break;
		}
		case HSV:
		{
			T r = this->r();
			T g = this->g();
			T b = this->b();
			T h = 0;
			T s = 0;
			T v = 0;
			T rgb_min = MIN3(r, g, b);
			T rgb_max = MAX3(r, g, b);

			if ((v = rgb_max) != 0)
			{
				r /= v;
				g /= v;
				b /= v;
				rgb_min = MIN3(r, g, b);
				rgb_max = MAX3(r, g, b);

				if ((s = rgb_max - rgb_min) != 0)
				{
					r = (r - rgb_min)/(rgb_max - rgb_min);
					g = (g - rgb_min)/(rgb_max - rgb_min);
					b = (b - rgb_min)/(rgb_max - rgb_min);
					rgb_min = MIN3(r, g, b);
					rgb_max = MAX3(r, g, b);

					if (rgb_max == r)
					{
						h = 0.0 + 60.0 * (g - b);
						if (h < 0.0) h += 360.0;
					}
					else if (rgb_max == g)
						h = 120.0 + 60.0 * (b - r);
					else /* rgb_max == rgb.b */
						h = 240.0 + 60.0 * (r - g);
				}
			}
			this->m_space  = HSV;
			this->m_data[0] = h;
			this->m_data[1] = s;
			this->m_data[2] = v;
			break;
		}
	}
}

template<typename T>
void
lenactions::pixel<T>::to_stream(std::ostream& out, format out_format)
{
	switch(out_format)
	{
		case P1:
			out << (int) (this->v() <= .5)   << " ";
			break;
		case P2:
			out << (int) (255 * this->v())   << " ";
			break;
		case P3:
			out << (int) (255 * this->r())   << " "
					<< (int) (255 * this->g())   << " "
					<< (int) (255 * this->b())   << " ";
			break;
		case P4:
			// out << (unsigned char) (this->v() <= .5);
			throw;
			break;
		case P5:
			out << (unsigned char) (255 * this->v());
			break;
		case P6:
			out << (unsigned char) (255 * this->r())
					<< (unsigned char) (255 * this->g())
					<< (unsigned char) (255 * this->b());
	}
}

template<typename T>
lenactions::pixel<T>
lenactions::pixel<T>::quadratic(pixel<T>& pi, pixel<T>& pj)
{
	T r = sqrt(pi.r()*pi.r() + pj.r()*pj.r());
	T g = sqrt(pi.g()*pi.g() + pj.g()*pj.g());
	T b = sqrt(pi.b()*pi.b() + pj.b()*pj.b());
	return pixel({r, g, b}, RGB);
}

template<typename T>
lenactions::pixel<T>
lenactions::pixel<T>::angleteinte(pixel<T>& pi, pixel<T>& pj)
{
	T h = atan2(pj.v(), pi.v()) * 180. / M_PI;
	T s = 1;
	T v = sqrt(pi.v()*pi.v()+pj.v()*pj.v());
	if (h < 0.) h += 360.;
	return pixel({h, s, v}, HSV);
}
