#include <iostream>
#include <fstream>
#include <vector>

#include <cstdio>
#include <cstdarg>
#include <cstring>
#include <ctime>

#include <lenactions/lenactions.hh>

// ----------------------------------------------------------------------------

#define CLR_NONE      -1
#define CLR_BLACK      0
#define CLR_RED        1
#define CLR_GREEN      2
#define CLR_YELLOW     3
#define CLR_BLUE       4
#define CLR_MAGENTA    5
#define CLR_CYAN       6
#define CLR_WHITE      7
#define ATTR_NONE      0
#define ATTR_BOLD      1
#define ATTR_DIM       2
#define ATTR_UNDERLINE 4
#define ATTR_BLINK     5
#define ATTR_REVERSE   7

struct timespec start;

void print(int fg, int attr, const char * format, ...)
{
	va_list args;
	va_start (args, format);

	struct timespec current;
	clock_gettime(CLOCK_REALTIME, &current);

	double dtime = (double) (current.tv_sec - start.tv_sec) +  (double) (current.tv_nsec - start.tv_nsec) / 1000000000;

	printf("[ %.3f ] ", dtime);
	if(fg == -1) printf("%c[%dm", 27, attr);
	else         printf("%c[%d;%dm",27,attr,(30+fg));
	vprintf(format, args);
	printf("%c[%dm\n", 27, 0);
	va_end (args);
}

// ----------------------------------------------------------------------------

char** arguments;

typedef struct
{
	std::string str;
	size_t      arg;
	size_t      opt;
	std::string arg_desc;
	std::string descriptor;
} option;
typedef std::vector<option> options;


int getopt(int argc, char* argv[], options opts)
{
	static size_t idx = 1;
	while (idx != (size_t) argc && argv[idx][0] != '-') idx++;
	if (idx == (size_t) argc) return -1;
	for (option o : opts)
	{
		if (o.str == &argv[idx][1])
		{
			if (idx + o.arg >= (size_t) argc)
			{
				std::cerr << "Option `" << argv[idx] << "` needs " << o.arg << " arguments" << std::endl;
				return -1;
			}
			if (arguments) delete[] arguments;
			arguments = new char*[o.arg];
			for (size_t j=0; j<o.arg; ++j) arguments[j] = argv[++idx];
			idx++;
			return o.opt;
		}
	}
	if(arguments) delete[] arguments;
	arguments = new char*[1];
	arguments[0] = argv[idx];
	idx++;
	return '?';
}

void help(options opts)
{
	printf("Usage: exec <commands>\n");
	printf("commands are:\n");
	for (option o : opts)
		printf("  -%-10s %-20s %s\n", o.str.c_str(), o.arg_desc.c_str(), o.descriptor.c_str());
}

int format_of_string(const char* str)
{
	if      (!strcmp(str, "P1") ) return lenactions::P1;
	else if (!strcmp(str, "P2") ) return lenactions::P2;
	else if (!strcmp(str, "P3") ) return lenactions::P3;
	else if (!strcmp(str, "P4") ) return lenactions::P4;
	else if (!strcmp(str, "P5") ) return lenactions::P5;
	else if (!strcmp(str, "P6") ) return lenactions::P6;
	print(CLR_RED, ATTR_BOLD, "Unknown format `%s`, export has been skipped", str);
	return -1;
}


const options listopts =
{
	{	"i",        1,   2, "<file>",         "Load image from file"                                          },
	{	"input",    1,   2, "<file>",         "Load image from file"                                          },
	{	"o",        2,   3, "<format><file>", "Save image to file using determined format"                    },
	{	"output",   2,   3, "<format><file>", "Save image to file using determined format"                    },
	{	"hist",     1,   5, "<file>",         "Export histogram data to file (gnuplot format)"                },
	{	"rotate",   1,  16, "<angle>",        "Rotate colors in HSV space by angle (degres)"                  },
	{	"gamma",    1,   6, "<gamma>",        "Gamma correction"                                              },
	{	"norm",     0,   7, "",               "Image normalization"                                           },
	{	"smooth",   0,   9, "",               "Image smoothing"                                               },
	{	"sobel1",   0,  10, "",               "Sobel contour detection"                                       },
	{	"sebel2",   0,  11, "",               "Sobel contour detection (theta value)"                         },
//{	"hide",     1,  14, "<file>",         "Hide `file` in current image, images must be of the same size" },
//{	"reveal",   0,  15, "",               "Reveal the image hidded in currently loaded image"             },
	{ "h",        0, 'h',	"",               "Display this list of options"                                  },
	{	"help",     0, 'h',	"",               "Display this list of options"                                  },
	{	"-help",    0, 'h',	"",               "Display this list of options"                                  }
};

using byte = double;
int main(int argc, char* argv[])
{

	srand (time(NULL));
	clock_gettime(CLOCK_REALTIME, &start);

	print(CLR_NONE, ATTR_UNDERLINE, "Starting execution");
	lenactions::image<byte> img;

	int opt;
	while ((opt = getopt(argc, argv, listopts)) != -1)
	{
		switch(opt)
		{
			case 2:	// INPUT
			{
				print(CLR_GREEN, ATTR_BOLD, "Importing picture from file `%s`", arguments[0]);
				std::ifstream file(arguments[0]);
				if(file.is_open()) img.from_stream(file);
				else
				{
					print(CLR_RED, ATTR_BOLD, "Could not open file");
					return 0;
				}
				break;
			}
			case 3:	// OUTPUT
			{
				int f;
				print(CLR_GREEN, ATTR_BOLD, "Exporting picture to file `%s` with format %s", arguments[0], arguments[1]);
				if ((f = format_of_string(arguments[0])) != -1)
				{
					std::ofstream file(arguments[1]);
					img.to_stream(file, (lenactions::format) f);
				}
				break;
			}
			case 16:
			{
				byte angle = atof(arguments[0]);
				print(CLR_NONE, ATTR_NONE, "Rotating HSV space by %.1f degres", angle);
				img.rotate(angle);
				break;
			}
			case 5:
			{
				print(CLR_NONE, ATTR_NONE, "Printing histogram");
				std::ofstream file(arguments[0]);
				img.histogram(file);
				break;
			}
			case 6:
			{
				byte y = atof(arguments[0]);
				print(CLR_NONE, ATTR_NONE, "Gamma correction (%d)", y);
				img.gamma(y);
				break;
			}
			case 7:
			{
				print(CLR_NONE, ATTR_NONE, "Image normalisation");
				img.normalize();
				break;
			}
			case 9:
				print(CLR_NONE, ATTR_NONE, "Smoothing");
				img.smooth();
				break;
			case 10:
				print(CLR_NONE, ATTR_NONE, "Sobel (quadratic merge)");
				img.contour_sobel(lenactions::pixel<byte>::quadratic);
				break;
			case 11:
				print(CLR_NONE, ATTR_NONE, "Sobel (theta coloring)");
				img.contour_sobel(lenactions::pixel<byte>::angleteinte);
				break;
			case 'h':
				help(listopts);
				break;
			case '?':
				print(CLR_YELLOW, ATTR_BOLD, "Option `%s` not recognized and skipped", arguments[0]);
				break;
		}
	}
	print(CLR_NONE, ATTR_UNDERLINE, "Execution finished");
	return 0;
}
